//gcc -Wall -S delay.c
//gcc -Wall -c delay.c && objdump -d delay.o > delay.obj && gcc -Wall -o delay delay.o && time delay 100000000

#include <stdlib.h>
#include <stdio.h>

int main (int argc, char *argv[]) {
//2147483648 is maximum!!!
unsigned long loops=strtoul(argv[1],NULL,10);
int d0; //dummy temporary variable used by assembler code (see =&a)
printf("loops %lu\n",loops);

__asm__ __volatile__(
	"\tjmp 1f\n"
	".align 16\n"
	"1:\tjmp 2f\n"
	".align 16\n"
	"2:\tdecl %0\n\tjns 2b"
	:"=&a" (d0)
	:"0" (loops));
printf("loops %lu\n",loops);
printf("do %d\n",d0);

return 0;
}
/*
 asm ( assembler template 
	: output operands                            
	: input operands                              
	: list of clobbered registers     
     );

Jump to label 1: jmp 1f
Jump to label 2: jmp 2f
decl %0 => decrement d0 variable (move loops to %eax, decrement %eax, result in %eax, if SF = 1 decrement %eax and finaly copy %eax to d0)
jns 2b => Jump if no Sign (>=0) (SF flag)
Sign Flag (SF) - set to 1 when result is negative.

:"=&a" (d0)
= => output, referenced as %0 in "assembler template"
a => use one of %eax, %ax, %al registers
& => contens of %eax will change before completing the function
"0" here specifies the same constraint as the 0th output variable. That is, it specifies that the output instance of var should be stored in %eax only.


http://www.ibiblio.org/gferg/ldp/GCC-Inline-Assembly-HOWTO.html#s3

The assembler template consists of assembly instructions. Each operand is described by an operand-constraint string followed by the C expression in parentheses. A colon separates the assembler template from the first output operand and another separates the last output operand from the first input, if any. Commas separate the operands within each group.


The assembler template contains the set of assembly instructions that gets inserted inside the C program. The format is like: either each instruction should be enclosed within double quotes, or the entire group of instructions should be within double quotes. Each instruction should also end with a delimiter. The valid delimiters are newline(\n) and semicolon(&semi;). ’\n’ may be followed by a tab(\t). We know the reason of newline/tab, right?. Operands corresponding to the C expressions are represented by %0, %1 ... etc.

 "constraint" (C expression) is the general form. For output operands an additional modifier will be there. Constraints are primarily used to decide the addressing modes for operands. They are also used in specifying the registers to be used.

If we use more than one operand, they are separated by comma.

In the assembler template, each operand is referenced by numbers. Numbering is done as follows. If there are a total of n operands (both input and output inclusive), then the first output operand is numbered 0, continuing in increasing order, and the last input operand is numbered n-1. The maximum number of operands is as we saw in the previous section. 

In some cases, a single variable may serve as both the input and the output operand. Such cases may be specified in "asm" by using matching constraints.

asm ("incl %0" :"=a"(var):"0"(var));

We saw similar examples in operands subsection also. In this example for matching constraints, the register %eax is used as both the input and the output variable. var input is read to %eax and updated %eax is stored in var again after increment. "0" here specifies the same constraint as the 0th output variable. That is, it specifies that the output instance of var should be stored in %eax only. This constraint can be used:

    * In cases where input is read from a variable or the variable is modified and modification is written back to the same variable.
    * In cases where separate instances of input and output operands are not necessary.

The most important effect of using matching restraints is that they lead to the efficient use of available registers. 

While using constraints, for more precise control over the effects of constraints, GCC provides us with constraint modifiers. Mostly used constraint modifiers are

   1. "=" : Means that this operand is write-only for this instruction; the previous value is discarded and replaced by output data.
   2. "&" : Means that this operand is an earlyclobber operand, which is modified before the instruction is finished using the input operands. Therefore, this operand may not lie in a register that is used as an input operand or as part of any memory address. An input operand can be tied to an earlyclobber operand if its only use as an input occurs before the early result is written.

      The list and explanation of constraints is by no means complete. Examples can give a better understanding of the use and usage of inline asm. In the next section we’ll see some examples, there we’ll find more about clobber-lists and constraints. 

Constraints "&S", "&D", "&a" say that the registers esi, edi and eax are early clobber registers, ie, their contents will change before the completion of the function.

http://webster.cs.ucr.edu/AoA/DOS/

*/
